#include<stdio.h>
#include<sys/time.h>

#define HALFSIZE 6
#define COUNT0 4
int LEVEL=8;

struct Board
{
  char hole[HALFSIZE*2];
  char gain[2];
  int flag ; //0: cote 0:5, 1: cote 6:11
};
typedef struct Board *board;

int timeMs()
{
  struct timeval tm;
  int t;
  
  gettimeofday(&tm,NULL);
  t=(tm.tv_sec*1000)+(tm.tv_usec/1000);
  return t;
}

void boardInit(board b)
{
  int i;
  for(i=0;i<HALFSIZE*2;i++) b->hole[i]=COUNT0;
  b->gain[0]=b->gain[1]=0;
  b->flag=0;
}

void boardCopy(board b,board a)
{
  int i;
  for(i=0;i<HALFSIZE*2;i++) b->hole[i]=a->hole[i];
  b->gain[0]=a->gain[0];
  b->gain[1]=a->gain[1];
  b->flag=a->flag;
}

int boardEnd(board b)
{
  int i;
  for(i=0;i<HALFSIZE*2;i++) if(b->hole[i]) return 0;
  return 1;
}

void boardPrint(board b)
{
  int i;
  printf(" ");
  for(i=0;i<HALFSIZE;i++) printf("  %c",i+65);
  printf("\n");

  if (b->flag) printf("*"); else printf(" ");
  for(i=HALFSIZE*2-1;i>=HALFSIZE;i--) printf(" %2d",b->hole[i]);
  printf(" / %d\n",b->gain[1]);

  if (!b->flag) printf("*"); else printf(" ");
  for(i=0;i<HALFSIZE;i++) printf(" %2d",b->hole[i]);
  printf(" / %d\n",b->gain[0]);
}


// on joue une position 0<=p<6
// on retourne 1 si ok, 0 si impossible
int boardPlay(board b,int p)
{
  int i;
  if (p<0) return 0;
  if (p>HALFSIZE) return 0;
  if (p==HALFSIZE)
    {
      p=b->flag?HALFSIZE:0;
      for(i=0;i<HALFSIZE;i++) if (b->hole[p+i]) return 0;
      for(i=0;i<HALFSIZE*2;i++)
	{
	  b->gain[b->flag]+=b->hole[i];
	  b->hole[i]=0;
	}
      b->flag=1-b->flag;
      return 1;
    }
  if (b->flag) p=HALFSIZE*2-1-p;
  i=b->hole[p];
  if (!i) return 0;
  b->hole[p]=0;
  while(i)
    {
      p++;
      if (p>=HALFSIZE*2) p=0;
      b->hole[p]++;
      i--;
    }
  if (b->hole[p]<=2)
    {
      if (((b->flag)&&(p<HALFSIZE))||((!b->flag)&&(p>=HALFSIZE)))
	{
	  while((b->hole[p])&&(b->hole[p]<=2))
	    {
	      b->gain[b->flag]+=b->hole[p];
	      b->hole[p]=0;
	      p--;
	      if (p<0) p=HALFSIZE*2-1;
	    }
	}
    }
  b->flag=1-b->flag;
  return 1;
}

int count;

// retourne le gain, vu de celui qui doit jouer
int boardEval(board b)
{
  int x=b->gain[b->flag]-b->gain[1-b->flag];
  count++;
  //  printf("eval %d\n",x);
  return x;
}

// algorithme de test de possibilite d'atteindre 'seuil' avec un niveau de profondeur 'level'
// le plateau est positionne pour celui qui doit jouer
int boardTest(board b0,int level,int seuil)
{
  struct Board b;
  int i,k;
 
  if (!level) return (boardEval(b0)>=seuil)?0:-1;
  boardCopy(&b,b0);
   for(i=0;i<=HALFSIZE;i++)
    {
      if (boardPlay(&b,i))
	{
	  k=boardTest(&b,level-1,seuil);
	  //          printf("play %d (%d)->%d\n",i,level,k);
	  if(k<0) return i;
	  boardCopy(&b,b0);
	}
    }
  return -1;
}

void boardDicho(board b0,int level,int smin,int smax,int* ablescore,int* ablemove)
{
  int k;
  int seuil=(smin+smax)/2;
  if (smin+1>=smax) return;
  //  printf("[%3d  -  %3d  -  %3d]\n",smin,seuil,smax);
  k=boardTest(b0,level,seuil);
  if (k<0) boardDicho(b0,level,smin,seuil,ablescore,ablemove);
  else
    {
      *ablescore=seuil;
      *ablemove=k;
      boardDicho(b0,level,seuil,smax,ablescore,ablemove);
    }
}

int boardSolve(board b0,int level)
{
  int ablescore=0;
  int ablemove=-1;
  int delta;
  count=0;

  printf("thinking...\n");
  delta=timeMs();
  boardDicho(b0,level*2,-COUNT0*HALFSIZE*2,COUNT0*HALFSIZE*2,&ablescore,&ablemove);
  delta=timeMs()-delta+1;
  printf("-> %c {%d} count=%d t=%dms (%d c/ms)\n",65+ablemove,ablescore,count,delta,count/delta);
  return ablemove;
}

// saisie d'un mouvement
int getmove()
{
  char buf[128];
  char m;

  printf("]");
  scanf("%s",buf);
  m=buf[0];
  if (m<'A') return -1;
  if (m<='A'+HALFSIZE) return m-'A';
  if (m<'a') return -1;
  if (m<='a'+HALFSIZE) return m-'a';
  return -1;
}

// boucle principale
int main(int argc,char** argv)
{
  int i,m;
  int n=0;
  int mode=2;
  int round=0;
  struct Board b0;
  i=0;
  printf("awelee - by Sylvain Huet - 2008\n");
  printf("-------------------------------\n");
  while(i<argc)
  {
	if (!strcmp(argv[i],"-h"))
	{
		printf("usage : awelee -l level -cc|-ch|-hc|-hh\n");
		return 0;
	}
	if ((!strcmp(argv[i],"-l"))&&(argc>i+1))
	{
		i++;
		LEVEL=atoi(argv[i]);
	}
	if (!strcmp(argv[i],"-cc")) mode=3;
        if (!strcmp(argv[i],"-ch")) mode=1;
        if (!strcmp(argv[i],"-hc")) mode=2;
        if (!strcmp(argv[i],"-hh")) mode=0;
	i++;
  }
  boardInit(&b0);

  printf("level=%d\n",LEVEL);
  printf("player1=%s\n",(mode&1)?"computer":"human");
  printf("player2=%s\n",(mode&2)?"computer":"human");

  while(!(boardEnd(&b0)))
    {
      printf("%d.\n",n++);
      boardPrint(&b0);
	if (mode&(1<<round))
	{
	      m=boardSolve(&b0,LEVEL);
	}
	else
	{
	      m=getmove();
	}
      m=boardPlay(&b0,m);
      if (m)
	{
	 printf("ok\n");
	 round=1-round;
	}
      else printf("erreur\n");
    }
  printf("%d.\n",n++);
  boardPrint(&b0);
  return 0;
}
